/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include <fstream>
#include <list>
#include "DnaSequence.h"
#include "Util.h"

using namespace std;

const GenomicPosition DnaSequence::notFound = invalidGenomicPosition;

int
DnaSequence::compare(GenomicPosition pos, GenomicPosition n, const char* s2) const
{
	const char* s1 = _s.c_str() + pos;
	for (; n; ++s1, ++s2, --n) {
		if (*s1 < *s2) return -1;
		if (*s1 > *s2) return 1;
	}
	return 0;

}


#if 0
static GenomicPosition
findNaive(const char* a, int64_t alen, const char* b, int64_t blen)
{
	for (int64_t i = alen - blen; i >= 0; --i) {
		int64_t j = blen - 1;
		const char* bp = b + j;
		const char* ap = a + i + j;
		for (; j >= 0 and *ap == *bp; --j, --bp, --ap);
		if (j < 0) return i;
	}
	return DnaSequence::notFound;
}


GenomicPosition
DnaSequence::find(const DnaSequence& other, int64_t thisFrom, int64_t thisLen, int64_t otherFrom, int64_t otherLen) const
{
	const char *a = _s.c_str();
	int64_t al = strlen(a);
	if (thisFrom < 0) thisFrom += al;
	if (thisFrom < 0 or thisFrom >= al) return notFound;
	a += thisFrom;
	al -= thisFrom;

	const char *b = other._s.c_str();
	int64_t bl = strlen(b);
	if (otherFrom < 0) otherFrom += bl;
	if (otherFrom < 0 or otherFrom >= bl) return 0;
	b += otherFrom;
	bl -= otherFrom;

	if (thisLen == maxLength or thisLen > al)
		thisLen = al;
	else if (thisLen < 0) {
		thisLen += al;
		if (thisLen < 0) return notFound;
	}
	if (otherLen == maxLength or otherLen > bl)
		otherLen = bl;
	else if (otherLen < 0) {
		otherLen += bl;
		if (otherLen < 0) return 0;
	}

	GenomicPosition ret = findNaive(a, thisLen, b, otherLen);
	return ret == notFound ? ret : ret + thisFrom;
}

 
#endif



// http://en.wikipedia.org/wiki/Boyer-Moore-Horspool_algorithm
//
DnaSequence::PreparedFindTarget::PreparedFindTarget(const DnaSequence& needle, int64_t from, int64_t length)
	: _needle(needle.c_str()), _nlen(needle.length())
{
	if (_nlen == 0) {
		_needle = NULL;
		return;
	}
	if (from < 0) from += _nlen;
	if (from < 0 or from >= _nlen) {
		_needle = NULL;
		return;
	}
	_needle += from;
	_nlen -= from;

	if (length == maxLength) {
		// _nlen == length already
	} else if (length < 0) {
		_nlen += length;
		if (_nlen <= 0) {
			_needle = NULL;
			return;
		}
	} else  if (length == 0) {
		_needle = NULL;
		return;
	} else if (length < _nlen) {
		_nlen = length;
	}


	for (int i = 0; i < 256; i++) {
		_badCharShift[i] = _nlen;
	}

	GenomicPosition last = _nlen - 1;
	for (GenomicPosition i = 0; i < last; ++i) {
		_badCharShift[(int) _needle[i]] = last - i;
	}
}


GenomicPosition
DnaSequence::find(const PreparedFindTarget& other, int64_t thisFrom, int64_t thisLen) const
{
	if (!other.valid()) return 0;

	const char *haystack = _s.c_str();
	int64_t hlen = _s.length();
	if (thisFrom < 0) thisFrom += hlen;
	if (thisFrom < 0 or thisFrom >= hlen) return notFound;
	haystack += thisFrom;
	hlen -= thisFrom;

	if (thisLen == maxLength or thisLen > hlen) {
	} else if (thisLen < 0) {
		hlen += thisLen;
		if (hlen <= 0) return notFound;
	} else if (thisLen == 0) {
		return notFound;
	} else {
		hlen = thisLen;
	}

	const char* start = haystack;
	const char* needle = other._needle;
	GenomicPosition last = other._nlen - 1;
	while ((GenomicPosition) hlen > last) {
		for (GenomicPosition i = last; haystack[i] == needle[i]; i--)
			if (i == 0) return haystack - start + thisFrom;

		GenomicPosition jump = other.badCharShift(haystack[last]);
		hlen -= jump;
		haystack += jump;
	}

	return notFound;
}


GenomicPosition
DnaSequence::find(const DnaSequence& other, int64_t thisFrom, int64_t thisLen, int64_t otherFrom, int64_t otherLen) const
{
	const char *needle = other._s.c_str();
	int64_t nlen = other._s.length();

	if (otherFrom < 0) otherFrom += nlen;
	if (otherFrom < 0 or otherFrom >= nlen) return 0;
	needle += otherFrom;
	nlen -= otherFrom;

	if (otherLen == maxLength or otherLen > nlen) {
	} else if (otherLen < 0) {
		nlen += otherLen;
		if (nlen <= 0) return 0;
	} else  if (otherLen == 0) {
		return 0;
	} else {
		nlen = otherLen;
	}

	const char *haystack = _s.c_str();
	int64_t hlen = _s.length();
	if (thisFrom < 0) thisFrom += hlen;
	if (thisFrom < 0 or thisFrom >= hlen) return notFound;
	haystack += thisFrom;
	hlen -= thisFrom;

	if (thisLen == maxLength or thisLen > hlen) {
	} else if (thisLen < 0) {
		hlen += thisLen;
		if (hlen <= 0) return notFound;
	} else if (thisLen == 0) {
		return notFound;
	} else {
		hlen = thisLen;
	}

	const char* start = haystack;
	GenomicPosition i;
	GenomicPosition badCharShift[256];
	GenomicPosition last = nlen - 1;

	for (i = 0; i < 256; i++) {
		badCharShift[i] = nlen;
	}

	for (i = 0; i < last; ++i) {
		badCharShift[(int) needle[i]] = last - i;
	}

	while (hlen >= nlen) {
		for (i = last; haystack[i] == needle[i]; i--)
			if (i == 0) return haystack - start + thisFrom;

		GenomicPosition jump = badCharShift[(int) haystack[last]];
		hlen -= jump;
		haystack += jump;
	}
	return notFound;
}


void
loadFastaFile(const char* filename, /* returned: */ Vector<DnaNamedSequence*>& l)
{
	ifstream ifile(filename, ios_base::in);
	if (ifile.fail()) {
		cerr << ": Can't open file " << filename << " in "
			<< __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ << endl;
		exit(1);
	}

	string line;
	string currentName;
	string currentSequence;
	list<DnaNamedSequence*> ls;
	while (getline(ifile, line)) {
		if (line[0] == '>') {
			if (currentName.length()) {
				ls.push_back(new DnaNamedSequence(currentName, currentSequence));
			}
			currentSequence.clear();
			size_t pos = line.find(' ');
			if (pos == string::npos) pos = line.find('\t');
			if (pos == string::npos) currentName = line.substr(1);
			else currentName = line.substr(1, pos-1);
		} else {
			currentSequence += line;
		}
	}
	if (currentName.length())
		ls.push_back(new DnaNamedSequence(currentName.c_str(), currentSequence.c_str()));

	l.resize(ls.size());
	list<DnaNamedSequence*>::iterator i;
	uint16_t j;
	for (i = ls.begin(), j = 0; i != ls.end(); ++i, ++j)
		l[j] = *i;
}


int nameBefore(const DnaNamedSequence& a, const DnaNamedSequence& b)
{
	return a.name() < b.name();
}
