/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include <fstream>
#include "GenomeFragmentArray.h"

namespace GenomeFragmentLocal {
class LowHigh 
{
public:
	uint64_t _low;
	uint64_t _high;
	LowHigh(uint64_t l, uint64_t h) : _low(l), _high(h) {}
	LowHigh() {}
//	LowHigh(const LowHigh& o) : _low(o._low), _high(o._high) {cdbg << "LowHigh copy " << _low << "," << _high << endl;}
	uint64_t middle() const { return (_low + _high) / 2; }
};
};


void
GenomeFragmentArray::sort(ostream* logFile)	// quicksort
{
	if (_number <= 1) return;
	time_t time0 = time(NULL);
	if (logFile) *logFile << Util::getLocalTime(time0) << " sorting genomic fragments...";

	uint64_t reportedHigh = _number-1;
	uint64_t maxStackDepth = 0;
	uint64_t numComparisons = 0;

	Stack<GenomeFragmentLocal::LowHigh> stack;
	stack.push(GenomeFragmentLocal::LowHigh(0, _number-1));

	while (!stack.isEmpty()) {
		if (stack.depth() > maxStackDepth) maxStackDepth = stack.depth();
		GenomeFragmentLocal::LowHigh lh = stack.pop();
		if (logFile and reportedHigh > lh._high and (reportedHigh - lh._high) * 100 >= _number) {
			*logFile << (100 - 100 * lh._high / _number) << "% ";
			reportedHigh = lh._high;
		}

		// trivial if only a few elements:
		if (lh._low + 4 >= lh._high) {
			while (lh._low < lh._high) {
				uint64_t min = lh._low;
				for (uint64_t i = lh._low + 1; i <= lh._high; ++i) {
					numComparisons++;
					if (compareF(min, i) > 0) min = i;
				}
				if (min != lh._low) swap(min, lh._low);
				lh._low++;
			}
			continue;
		}

		// partition:

		// first put the median of (low, mid, high) at the middle as a pivot:
		const uint64_t mid = lh.middle();
		numComparisons++;
		if (compareF(mid, lh._low) < 0) swap(mid, lh._low);
		numComparisons++;
		if (compareF(mid, lh._high) > 0) {
			swap(mid, lh._high);
			numComparisons++;
			if (compareF(mid, lh._low) < 0) swap(mid, lh._low);
		}

		const GenomicPosition pivotLoc(_array[mid]);
		const char * const pivotFrag = fragment(mid);
		// at this stage _array[lh._low] < pivotValue and _array[lh._high] > pivotValue
		_array[mid] = _array[lh._high-1];

//#define SORT_METHOD_1

#ifdef SORT_METHOD_1 // method 1:

		uint64_t low = lh._low+1;
		uint64_t high = lh._high-2;
		while (low < high) {
			while (low < high and (++numComparisons, compareF(low, pivotFrag) < 0)) low++;
			while (low < high and (++numComparisons, compareF(high, pivotFrag) > 0)) high--;
			if (low < high) {
				swap(low, high);
				low++;
				high--;
			}
		}
		numComparisons++;
		const uint64_t pivot = compareF(low, pivotFrag) < 0 ? low + 1 : low;

// end method 1
#else // method 2:
		uint64_t pivot = lh._low + 1;
		for (uint64_t i = lh._low + 1; i < lh._high - 1; ++i) {
			numComparisons++;
			if (compareF(i, pivotFrag) < 0) {
				if (i != pivot) swap(i, pivot);
				pivot++;
			}
		}
#endif // end method 2
		if (pivot > lh._low + 1)
			stack.push(GenomeFragmentLocal::LowHigh(lh._low, pivot-1));

		if (pivot + 1 < lh._high) {
			_array[lh._high-1] = _array[pivot];
			stack.push(GenomeFragmentLocal::LowHigh(pivot+1, lh._high));
		}
		_array[pivot] = pivotLoc;
	}

	if (logFile) *logFile << " done in " << Util::hms(time(NULL)-time0)
		<< " max stack depth " << maxStackDepth 
		<< " number of comparisons " << numComparisons
		<< " number of elements " << _number
		<< " N log(N) " <<  _number * (uint64_t) log2(_number)
		<< endl;
}


GenomeFragmentArray::GenomeFragmentArray(const Genome& genome, Strand strand, ostream* logFile, GenomicPosition fragmentLength, bool writeCache)
	: _genome(genome), _genomeLength(genome.length()), _fragmentLength(fragmentLength), _strand(strand), _array(NULL), _sequence(NULL), _number(0)
{
	if (loadCache()) return;

	const ChromosomeIndex numChromo = _genome.numberChromosomes();
	if (!numChromo) return;

	if (logFile) *logFile << Util::getLocalTime(time(NULL)) << " building list of fragments...";

	_sequence = new char [ ((size_t) _genomeLength) * 2 + 2 ];
	const char* genSequence = _genome.sequence(_strand);
	strcpy(_sequence, genSequence);
	_sequence[_genomeLength] = '$';
	strcpy(_sequence + _genomeLength + 1, genSequence);

	_array = new GenomicPosition [ _genomeLength + 1 ];

	for (GenomicPosition pos = 0; pos <= _genomeLength; ++pos, ++_number)
		_array[_number] = pos;

	if (logFile) *logFile << endl;

	sort(logFile);

/// check array is sorted:
#ifdef DEBUG
	if (logFile) {
		uint64_t numNonUnique = 0;
		uint64_t maxNumCopies = 0;
		const GenomicPosition len = _fragmentLength ? _fragmentLength : 100;

		*logFile << "Verifying sorted genome: ";
//		cdbg << String(fragment(0), len) << endl;
		for (uint64_t i = 1; i < _number; ++i) {
//			cdbg << String(fragment(i), len) << endl;
			int cmp = compareF(i, i - 1);
			if (cmp < 0) {
				*logFile << "not sorted: " << i << ":" << String(fragment(i-1), len) << " vs " << String(fragment(i), len) << endl;
			} else if (cmp == 0) {
				numNonUnique++;
				uint64_t j, numCopies = 2;
				for (j = i + 1; j < _number; ++j) {
					int cmp = compareF(i, j);
					if (cmp > 0) {
						*logFile << "not sorted: " << i << ", " << j << " :" << String(fragment(i), len) << " vs " << String(fragment(j), len) << endl;
					} else if (cmp == 0) {
						numCopies++;
					} else {
						break;
					}
				}
				if (numCopies > maxNumCopies) maxNumCopies = numCopies;
				i = j - 1;
			}
		}
		*logFile << " number of non-unique fragments " << numNonUnique << ", highest number of copies " << maxNumCopies << endl;
	}
#endif

	if (writeCache) saveCache();

	delete [] _sequence;
	_sequence = NULL;
}

static const char* const cacheSignature = "PrInSeS genome cache";
static const char* const cacheVersion = "v1";


void
GenomeFragmentArray::saveCache() const
{
	String filename = cacheFilename();
	cerr << "Writing cache file... " << filename << "\n";
	ofstream fh(filename.c_str(), ios_base::out|ios_base::binary);
	if (fh.fail()) {
		cerr << "Warning: Can't write genome cache in " << filename << "\n";
		return;
	}
	fh << cacheSignature
		<< '\t' << cacheVersion
		<< '\t' << _number
		<< '\t' << _genome.checksum()
		<< '\t' << Util::getLocalTime(time(NULL))
		<< endl << flush;

	fh.write((const char*) _array, _number * sizeof(_array[0]));
}


bool
GenomeFragmentArray::loadCache()
{
	String filename = cacheFilename();
	ifstream fh(filename.c_str(), ios_base::in|ios_base::binary);
	if (fh.fail()) return false;

	cerr << "Loading cached fragments from file " << filename << "... ";
	String header;
	getline(fh, header);
	Vector<String> fields;
	Util::split(header, '\t', fields);
	uint32_t checksum;
	if (fields.dimension() != 5 or fields[0] != cacheSignature or fields[1] != cacheVersion
	or !Util::value<uint64_t>(_number, fields[2]) or !Util::value<uint32_t>(checksum, fields[3])) {
		_number = 0;
		cerr << "Warning: Unrecognised cache file format for " << filename << endl;
		return false;
	}

	if (checksum != _genome.checksum()) {
		_number = 0;
		cerr << "Warning: Out of date cache file " << filename << endl;
		return false;
	}

	_array = new GenomicPosition[_number];
	fh.read((char*) _array, _number * sizeof(_array[0]));
	if (fh.fail() or fh.eof()) {
		cerr << "Warning: Problem reading cache file " << filename << endl;
		reset();
		return false;
	}

	cerr << "done" << endl;
	return true;
}


const static int32_t mapSize = 16777216; // 24-bit index, see also index() method


GenomeFragmentArrayIterator
GenomeFragmentArray::findFragment(const char* frag, GenomicPosition len) const
{
	uint64_t low = 0, high = _number;
	int cmp = -1;
	if (_fragmentLength and len > _fragmentLength) len = _fragmentLength;

//	String b = String(frag, len);
	uint64_t current;
	while (low < high) {
		current = (low + high) / 2;
//		String a = String(_genome.sequence(_strand) + _array[current], len);
		cmp = compare(current, frag, len, false);
//		cdbg << "cmp " << current << " " << a << " with " << b << " (" << a.compare(b) << ") result=" << cmp << endl;
		if (cmp > 0) {
			high = current;
		} else if (cmp < 0) {
			low = current + 1;
		} else {	// cmp == 0
			break;
		}
	}

	if (cmp) return GenomeFragmentArrayIterator(*this);	// no reads found

	for (high = current + 1; high < _number and compare(high, frag, len, false) == 0; ++high) {}
	high--;
	for (low = current; low > 0 and compare(low-1, frag, len, false) == 0; --low) {}

	return GenomeFragmentArrayIterator(*this, low, high);
}


int
GenomeFragmentArray::compare(uint64_t i, const char* frag, GenomicPosition len, bool wrapAround) const
{
	if (_fragmentLength)
		assert(len <= _fragmentLength);
	else
		assert(len <= _genomeLength + 1);

	const char* genSequence = _genome.sequence(_strand);
	GenomicPosition pos = _array[i];
	if (_strand.isNegative() and pos != _genomeLength) pos = _genomeLength - pos - 1;

	for (; len and pos < _genomeLength; ++pos, ++frag, --len) {
		if (genSequence[pos] != *frag)
			return genSequence[pos] < *frag ? -1 : 1;
	}
	if (!len) return 0;
	if (!wrapAround or *frag != '$') return -1;

	// wrap around and continue:
	for (frag++, len--, pos = 0; len; ++pos, ++frag, --len) {
		if (genSequence[pos] != *frag)
			return genSequence[pos] < *frag ? -1 : 1;
	}
	return 0;
}


int
GenomeFragmentArray::compare(uint64_t i, const GenomeFragmentArray& other, uint64_t j, GenomicPosition len, bool wrapAround, /* returned: */ GenomicPosition& numberSameBases) const
{
	const char* seqi = _genome.sequence(_strand);
	const char* seqj = other._genome.sequence(other._strand);
	GenomicPosition posi = _array[i], posj = other._array[j];
	const GenomicPosition gli = _genomeLength;
	const GenomicPosition glj = other._genomeLength;
	if (len == 0 or len > gli + 1 or len > glj + 1) len = min(gli, glj) + 1;
	numberSameBases = 0;

	if (_strand.isNegative() and posi < gli) posi = gli - posi - 1;
	if (other._strand.isNegative() and posj < glj) posj = glj - posj - 1;

	for (; len and posi < gli and posj < glj; ++posi, ++posj, --len, ++numberSameBases) {
		if (seqi[posi] != seqj[posj])
			return seqi[posi] < seqj[posj] ? -1 : 1;
	}
	if (!len) return 0;
	if (posi < gli) return 1;
	if (posj < glj) return -1;
	if (!wrapAround) return 0;
	seqi = _genome.sequence(_strand);
	seqj = other._genome.sequence(other._strand);
	for (len--, numberSameBases++; len; ++seqi, ++seqj, --len, ++numberSameBases) {
		if (*seqi != *seqj)
			return *seqi < *seqj ? -1 : 1;
	}
	return 0;
}


void
GenomeFragmentArray::output(ostream& o) const
{
	if (!_array) return;
	const char* seq = _genome.sequence(_strand);
	const GenomicPosition len = _genomeLength;

	o << "Genome fragment array:" << endl << "index\tposition\tgenome" << endl;
	for (uint64_t i = 0; i < _number; ++i) {
		GenomicPosition pos = _array[i];
		if (_strand.isNegative() and pos < len) pos = len - pos - 1;
		o << i << '\t' << pos << '\t';
		GenomicPosition l = len + 1;
		for (; pos < len and l; l--, pos++) {
			o << seq[pos];
		}
		o << '$'; l--;
		for (pos = 0; l; l--, pos++) {
			o << seq[pos];
		}
		o << endl;
	}
	o << "Genome fragment array end" << endl;
}
