/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef GENOMEFRAGMENTARRAY_H
#define GENOMEFRAGMENTARRAY_H

#include "Genome.h"

class GenomeFragmentArrayIterator;

class GenomeFragmentArray
{
public:
	GenomeFragmentArray(const Genome& genome, Strand strand, ostream* logFile = NULL, GenomicPosition fragmentLength = 0, bool writeCache = true);
	~GenomeFragmentArray() { reset(); }
	uint64_t number() const { return _number; }

	// finds all fragments whose first n bases are the same as those of "frag"
	GenomeFragmentArrayIterator findFragment(const char* frag, GenomicPosition n) const;
	// returns all fragments starting with the same sequence as "frag"
	GenomeFragmentArrayIterator findFragment(const char* frag) const;

	int compare(uint64_t i, const char* frag, GenomicPosition len, bool wrapAround) const;

	// fistDifference is set to the first position where this.fragment[i] differs from other.fragment[j]:
	// use len = 0 for unlimited
	int compare(uint64_t i, const GenomeFragmentArray& other, uint64_t j, GenomicPosition len, bool wrapAround, /* returned: */ GenomicPosition& numberSameBases) const;

	// returns true if "frag" matches (the beginning of) one of the short sequences
	bool contains(const char* frag) const ;

	const Genome& genome() const { return _genome; }
	GenomicPosition operator[](uint64_t i) const { return _array[i]; }

	// returns the j'th character of fragment i
	char fragmentBase(uint64_t i, GenomicPosition j) const {
		const GenomicPosition n = _fragmentLength ? _fragmentLength - 1 : _genomeLength;
		assert(j <= n);
		GenomicPosition pos = _array[i];
		if (pos > numeric_limits<GenomicPosition>::max() - j)
			return _genome.sequence(_strand)[pos + j - n - 1];
		pos += j;
		if (pos < n) return _genome.sequence(_strand)[pos];
		if (pos == n) return '$';
		return _genome.sequence(_strand)[pos - n - 1];
	
	}
	void reset() {
		if (_array) { delete [] _array; _array = NULL; }
		if (_sequence) { delete [] _sequence; _sequence = NULL; }
		_number = 0;
	}

	void output(ostream&) const;

private:
	// the following two are not implemented
	GenomeFragmentArray(const GenomeFragmentArray&);
	GenomeFragmentArray& operator=(const GenomeFragmentArray&);

	void swap(uint64_t i, uint64_t j) {
		GenomicPosition tmp(_array[i]); _array[i] = _array[j]; _array[j] = tmp;
	}
	const char* fragment(uint64_t i) const {
		GenomicPosition p = _array[i];
		if (_strand.isPositive() or p == _genomeLength) return _sequence + p;
		return  _sequence + (_genomeLength - p - 1);
	}
	int compareF(uint64_t i, uint64_t j) {
		return strncmp(fragment(i), fragment(j), _fragmentLength ? _fragmentLength : _genomeLength + 1);
	}
	int compareF(uint64_t i, const char* frag) const {
		return strncmp(fragment(i), frag, _fragmentLength ? _fragmentLength : _genomeLength + 1);
	}

	void sort(ostream* logfile);
	void saveCache() const;
	bool loadCache();
	String cacheFilename() const {
		stringstream filename;
		filename << _genome.filename();
		if (_fragmentLength) filename << "." << _fragmentLength;
		filename << '.' << (_strand.isPositive() ? "pos" : "neg") << ".prinses.cache";
		return filename.str();
	}

	const Genome& _genome;
	const GenomicPosition _genomeLength;
	const GenomicPosition _fragmentLength;
	Strand _strand;
	GenomicPosition* _array;
	char* _sequence;
	uint64_t _number;
};


class GenomeFragmentArrayIterator
{
public:
	GenomeFragmentArrayIterator(const GenomeFragmentArray& collection) : _collection(collection), _current(1), _last(0) {}
	GenomeFragmentArrayIterator(const GenomeFragmentArrayIterator& other) : _collection(other._collection), _current(other._current), _last(other._last) {}
	GenomeFragmentArrayIterator(const GenomeFragmentArray& collection, uint64_t from, uint64_t to) : _collection(collection), _current(from), _last(to) {}
	bool next(GenomicPosition& r) { 
		if (_current <= _last) {
			r = _collection[_current];
			_current++;
			return true;
		}
		return false;
	}
	uint64_t numberRemaining() const { return _current <= _last ? _last - _current + 1 : 0; }

private:
	const GenomeFragmentArray& _collection;
	uint64_t _current;
	uint64_t _last;
};

inline GenomeFragmentArrayIterator 
GenomeFragmentArray::findFragment(const char* frag) const { return findFragment(frag, strlen(frag)); }


#endif // GENOMEFRAGMENTARRAY_H
