/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef PRINSES_H
#define PRINSES_H

#include <vector>
#include <map>
#include <list>
#include "DnaSequence.h"
#include "Reads.h"
#include "Genome.h"
#include "GenomeFragmentArray.h"

namespace Prinses {

typedef enum { END_TERMINATOR_FOUND, END_LOW_COVER, END_MAX_LENGTH, END_REPEATS, END_GENOME_REPEAT } AssemblyEndReason;
extern const char assemblyEndReasonChar[]; //  = { 'f', 'l', 'm', 'r', 'g' };
inline ostream& operator<<(ostream& o, AssemblyEndReason s) { return o << assemblyEndReasonChar[s]; }

class AssemblyResult
{
public:
	DnaSequence _contig;	// without primer and terminator
	AssemblyEndReason _assemblyEndReason;
	int16_t _terminatorIndex;	// index to the terminators vector, or -1 if no terminator found
	uint32_t _score;

	AssemblyResult(const DnaSequence& contig, AssemblyEndReason aer, uint32_t score) : _contig(contig), _assemblyEndReason(aer), _terminatorIndex(-1), _score(score) {}
};
inline ostream& operator<<(ostream &o, const AssemblyResult& r) { return o << r._contig << " " << assemblyEndReasonChar[r._assemblyEndReason] << " " << r._terminatorIndex << " " << r._score; }

class AssemblyParameters
{
public:
	// input parameters:
	ostream* _viewFile;
	bool _verbose;
	bool _useScores;
	String _assemblyId;
	int _maxBranches;
	DnaSequence _primer;
	const ShortSequenceSortedArray& _reads;
	uint16_t _minDepth;
	uint16_t _maxDepth;
	uint16_t _minMatchLength;
	GenomicPosition _maxAssemblyLength;
	Vector<const char*> _terminators;
	uint16_t _terminatorLength;
	const GenomeFragmentArray& _genomeFragmentsPositive;
	const GenomeFragmentArray& _genomeFragmentsNegative;

	AssemblyParameters(ostream* viewFile, bool verbose, bool useScores,
		int maxBranches, const ShortSequenceSortedArray& reads, 
		uint16_t minDepth, uint16_t maxDepth,
		uint16_t minMatchLength, GenomicPosition maxAssemblyLength,
		uint16_t numTerminators, uint16_t terminatorLength,
		const GenomeFragmentArray& pos, const GenomeFragmentArray& neg);
};


void assemble(AssemblyParameters&, list<AssemblyResult>& results);

/* maps reads to refseq, returns number of reads used */
uint64_t mapReads(const char* refseq, GenomicPosition refseqLength, const ShortSequenceSortedArray& reads, const GenomeFragmentArray&,
	uint16_t minMatchLength, /* returned: */ float*& coverage, ostream* logFile = NULL);

class Alignment
{
public:
	GenomicPosition _position;
	uint32_t _negStrand:1, _numMismatches:15;
};

/* maps a read to reference genome and returns the number of hits */
/* returns -1 for multiple alignments, 0 for no alignments and 1 for one alignment (saved in "alignment") */
int mapSingleRead(const char* readseq, const GenomicPosition length, const GenomeFragmentArray& genomeFragments, /* returned: */ Alignment& alignment);

uint16_t loadMaqProfile(ifstream&, Vector<DnaNamedSequence*>& maqConsensus, map<String, const uint8_t *>* maqReadDepth = NULL);

}; // namespace Prinses

#endif
