/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef READS_H
#define READS_H

#include "ShortDnaBinarySequence.h"
#include "Genome.h"

// The following results in reads being indexed for faster retrieval. It uses an extra 128MB
#define READS_INDEXING

extern const char* prinsesReadFileSignature; // = "PrInSeS-formatted reads";
extern const char* prinsesReadFileVersion; // = "v2";

class SequencingRead
{
public:
	SequencingRead() {}
	uint8_t score(uint64_t i) const {
		return (i % 2 == 0) ? _score[i/2] / 16 : _score[i/2] & 15;
	}

	uint32_t count() const { return _count; }
	void set (const ShortDnaBinarySequence* seq, uint32_t count, const uint8_t* score) { _sequence = seq; _count = count; _score = score; }
	void set (const ShortDnaBinarySequence* seq, uint32_t count) { _sequence = seq; _count = count; }
	Base operator[](uint16_t i) const { return (*_sequence)[i]; }
	const ShortDnaBinarySequence *sequence() const { return _sequence; }
	const uint16_t length() const { return _sequence->length(); }
private:
	const ShortDnaBinarySequence* _sequence;
	uint32_t _count;
	const uint8_t* _score;
};

class ShortSequenceSortedArrayIterator;

class ShortSequenceSortedArray
{
public:
	ShortSequenceSortedArray(uint64_t number, uint16_t readLength, bool hasScores);
	~ShortSequenceSortedArray() {
			delete [] _memory;
			if (_map) delete [] _map;
	}
	uint64_t number() const { return _number; }
	uint64_t totalNumber() const { return _totalNumber; }

	// returns the Reads and the read length
	static ShortSequenceSortedArray* loadReads(istream&);

	// this takes a "genome" and produces a list of sequence of length "length" that are non-unique in the genome, these sequences are NOT reads, but the same class is used.
	static ShortSequenceSortedArray* nonUniqueShortSequences(const Genome& genome, uint16_t fragLength, ostream* logFile = NULL);
	static ShortSequenceSortedArray* nonUniqueShortSequences(const DnaSequence&, uint16_t fragLength, ostream* logFile = NULL);

	// returns all reads starting with the same sequence as "frag" and are longer than frag
	ShortSequenceSortedArrayIterator findFragment(const ShortDnaBinarySequence* frag) const;
	// finds all reads whose first n bases are the same as those of "frag"
	ShortSequenceSortedArrayIterator findFragment(const ShortDnaBinarySequence* frag, uint16_t n) const;

	// returns true if "frag" matches (the beginning of) one of the short sequences
	bool contains(const ShortDnaBinarySequence* frag) const ;

	const ShortDnaBinarySequence* sequence(uint64_t i) const { return (ShortDnaBinarySequence*) (_memory + i * _recordSize); }
	uint32_t count(uint64_t i) const { return *( (uint32_t*) (_memory + i * _recordSize + _readSize)); }
	const uint8_t* score(uint64_t i) const { return (uint8_t*) (_memory + i * _recordSize + _readSize + 4); }
	uint16_t readLength() const { return _readLength; }
	uint64_t memoryUsed() const { return _next - _memory; }
	const char* memory() const { return (char*) _memory; }
	bool hasScores() const { return _scoreSize != 0; }

private:
	// the following two are not implemented
	ShortSequenceSortedArray(const ShortSequenceSortedArray&);
	ShortSequenceSortedArray& operator=(const ShortSequenceSortedArray&);

	static ShortSequenceSortedArray* nonUniqueShortSequences(uint8_t*, uint64_t, uint16_t, ostream*);

	void add(const ShortDnaBinarySequence* s, uint32_t numCopies);

	// For prinses read files version 1
	void addV1(const String& seq, uint32_t count, const String& score);
	void addV1(const String& seq, uint32_t count);
	static ShortSequenceSortedArray* loadReadsV1(istream&);
	static int32_t index(const ShortDnaBinarySequence* seq) {
		const uint8_t *rawSeq = seq->rawSeq();
		return rawSeq[0] * 65536 + rawSeq[1] * 256 + rawSeq[2];
	}
	void buildMap();

	const uint8_t* _memory;
	uint8_t* _next;
	uint16_t _readLength;
	uint16_t _readSize;
	uint16_t _scoreSize;
	uint16_t _recordSize;
	uint64_t _number;	// number of unique sequences (reads)
	uint64_t _totalNumber;
	uint64_t* _map;		// maps first few bases of reads for quick search
};

class ShortSequenceSortedArrayIterator {
public:
	ShortSequenceSortedArrayIterator(const ShortSequenceSortedArray& collection) : _collection(collection), _current(1), _last(0) {}
	ShortSequenceSortedArrayIterator(const ShortSequenceSortedArrayIterator& other) : _collection(other._collection), _current(other._current), _last(other._last) {}
	ShortSequenceSortedArrayIterator(const ShortSequenceSortedArray& collection, uint64_t from, uint64_t to) : _collection(collection), _current(from), _last(to) {}
	bool next(SequencingRead& r) { 
		if (_current <= _last) {
			r.set(_collection.sequence(_current), _collection.count(_current), _collection.score(_current));
			_current++;
			return true;
		}
		return false;
	}
	uint64_t numberRemaining() const { return _current <= _last ? _last - _current + 1 : 0; }

private:
	const ShortSequenceSortedArray& _collection;
	uint64_t _current;
	uint64_t _last;
};


#endif
