/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef SHORTDNABINARYSEQUENCE_H
#define SHORTDNABINARYSEQUENCE_H

#include "DnaSequence.h"

// This class has variable size in memory, and should thus not be instantiated in the stack
// It is most appropriate to hold a sequencing read or other DNA sequence up to 65535 bp
// Each DNA base takes 2 bits of memory, so there it be only A, C, G, or T.
class ShortDnaBinaryBaseSequence
{
public:
	static uint16_t sequenceByteSize(uint16_t seqLength) {  return (seqLength + 3) / 4; }
	const uint8_t* rawSeq() const { return _sequence; }
	Base operator[](uint16_t i) const { return (Base) ((_sequence[i/4] >> (2*(3-i%4))) & 3); }
	void copy(const ShortDnaBinaryBaseSequence& from, uint16_t n);
	uint16_t set(const char*);	// from text sequence, returns number of bases or 0 if unsuccessful
	uint16_t set(const char*, uint16_t n);
	int compare(const ShortDnaBinaryBaseSequence& other, uint16_t n) const;
	bool equals(const ShortDnaBinaryBaseSequence& other, uint16_t n) const;

private:
	uint8_t _sequence[];
};


class ShortDnaBinarySequence
{
public:
	ShortDnaBinarySequence() : _length(0) {}
	ShortDnaBinarySequence(const ShortDnaBinarySequence* other) { copy(other); }
	static uint16_t memorySize(uint16_t seqLength) {  return sizeof(ShortDnaBinarySequence) + ShortDnaBinaryBaseSequence::sequenceByteSize(seqLength); }
	bool set(const char* s) { _length = _baseSeq.set(s); return _length > 0; }
	bool set(const char* s, uint16_t n) { _length = _baseSeq.set(s, n); return _length > 0; }
	void copy(const ShortDnaBinarySequence* from) { memcpy(this, from, memorySize(from->_length)); }
	void copy(const ShortDnaBinarySequence* from, uint16_t n) {
		if (n > from->_length)
			copy(from);
		else {
			_baseSeq.copy(from->_baseSeq, n);
			_length = n;
		}
	}
	uint16_t length() const { return _length; }
	uint16_t memorySize() const { return memorySize(_length); }

	// in the following, we are comparing the first n bases of each sequence:
	int compare(const ShortDnaBinarySequence* other, uint16_t n = 65535) const {
		uint16_t len = n;
		uint16_t len1 = _length;
		if (len > len1) len = len1;
		uint16_t len2 = other->_length;
		if (len > len2) len = len2;

		int cmp = _baseSeq.compare(other->_baseSeq, len);
		if (cmp) return cmp;
		if (n > len1 or n > len2) {
			if (len1 < len2) return -1;
			if (len1 > len2) return 1;
		}
		return 0;
	}
	bool equals(const ShortDnaBinarySequence* other, uint16_t n) const
	{
		uint16_t len1 = _length;
		uint16_t len2 = other->_length;
		if (len1 > n) len1 = n;
		if (len2 > n) len2 = n;
		if (len1 != len2) return false;

		return _baseSeq.equals(other->_baseSeq, len1);
	}
	bool equals(const ShortDnaBinarySequence* other) const { 
		if (_length != other->_length) return false;
		return memcmp(&this->_baseSeq, &other->_baseSeq, ShortDnaBinaryBaseSequence::sequenceByteSize(_length)) == 0;
	}
	Base operator[](uint16_t i) const { return _baseSeq[i]; }
	String toText() const;
	const uint8_t* rawSeq() const { return _baseSeq.rawSeq(); }

	void *operator new(size_t s, uint16_t seqLength) { return SimpleAllocator::allocate(memorySize(seqLength)); }
	void operator delete(void *p) { SimpleAllocator::free((uint8_t*) p); }

private:
	uint16_t _length;
	ShortDnaBinaryBaseSequence _baseSeq;		// variable size

};

inline ostream& operator<<(ostream& o, const ShortDnaBinarySequence& s) { return o << s.toText(); }
inline ostream& operator<<(ostream& o, const ShortDnaBinarySequence*s) { return o << s->toText(); }

class ShortDnaSequenceNode
{
	ShortDnaSequenceNode(const char* sequence) : _left(NULL), _right(NULL), _numCopies(1) { _seq.set(sequence); }
	friend class ShortDnaSequenceTree;

private:
	void *operator new(size_t s, uint16_t seqLength) { return new char [sizeof(ShortDnaSequenceNode) - sizeof(ShortDnaBinarySequence) + ShortDnaBinarySequence::memorySize(seqLength)]; }
	void operator delete(void *p) { delete [] (char*) p; }

	ShortDnaSequenceNode* _left;
	ShortDnaSequenceNode* _right;
	uint32_t _numCopies;
	ShortDnaBinarySequence _seq;
};


class ShortDnaSequenceTree
{
public:
	ShortDnaSequenceTree() : _root(NULL), _number(0), _numUnique(0), _maxDepth(0) {}
	~ShortDnaSequenceTree();
	uint32_t addNode(const char* sequence);	// returns the number of occurences of "sequence" in the tree
	bool contains(const char* sequence) const;

	uint16_t add(const char *sequence);
	uint64_t number() const { return _number; }
	uint64_t maxDepth() const { return _maxDepth; }
;
private:
	// no copying:
	ShortDnaSequenceTree(ShortDnaSequenceTree&);
	ShortDnaSequenceTree& operator=(ShortDnaSequenceTree&);

	ShortDnaSequenceNode* _root;
	uint64_t _number;
	uint64_t _numUnique;
	uint64_t _maxDepth;
};

#endif

