/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef UTIL_H
#define UTIL_H

#include <cstdlib>
#include <string>
#include <cstring>
#include <sstream>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <map>

#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

using namespace std;

extern ostream& cdbg;

template <class T> class Vector;

typedef uint32_t GenomicPosition;		// 32-bit unsigned integer may be too short for some genomes, may need to increase
extern const GenomicPosition invalidGenomicPosition; // = numeric_limits<GenomicPosition>::max();

#ifdef DEBUGxxxxx

// Naive implementation
// This helps when running in a debugger, stl version "string" makes it hard to see what's in there!
class String
{
public:
	String() : _s(NULL), _length(0) {}
	String(const char* s) : _s(NULL), _length(0) {
		if (!s) return;
		_length = strlen(s);
		if (!_length) return;
		_s = new char[_length+1];
		strcpy(_s, s);
	}
	String(const char* s, size_t n) : _s(NULL), _length(0) {
		if (!s) return;
		_length = strlen(s);
		if (n < _length) _length = n;
		if (!_length) return;
		_s = new char[_length+1];
		strncpy(_s, s, _length);
		_s[_length] = 0;
	}
	String(const string& s) : _s(NULL), _length(0) {
		assign(s.c_str(), s.length());
	}
	String(const String& other) : _s(NULL), _length(other._length) {
		if (_length) {
			_s = new char[_length+1];
			strcpy(_s, other._s);
		}
	}
	String(const String& other, size_t pos, size_t n = string::npos) : _s(NULL), _length(0) {
		const size_t l = other.length();
		if (pos >= l) return;
		if (n == string::npos or pos + n > l)
			n = l - pos;
		_length = n;
		_s = new char[n+1];
		strncpy(_s, other._s + pos, n);
		_s[n] = 0;
	}
	String(size_t n, char c = ' ') : _s(NULL), _length(n) {
		if (!n) return;
		_s = new char[n + 1];
		_s[n] = 0;
		for (; n; n--) _s[n-1] = c;
	}
	~String() { if (_s) delete [] _s; }
	size_t length() const { return _length; }
	const String& operator=(const String& other) {
		if (_s) delete [] _s;
		_length = other._length;
		if (_length) {
			_s = new char[_length+1];
			strcpy(_s, other._s);
		} else {
			_s = NULL;
		}
		return *this;
	}
	const String& operator=(const char* other) {
		if (_s) delete [] _s;
		_length = other ? strlen(other) : 0;
		if (_length) {
			_s = new char[_length+1];
			strcpy(_s, other);
		} else {
			_s = NULL;
		}
		return *this;
	}
	const String& operator=(const string& other) {
		if (_s) delete [] _s;
		_length = other.length();
		if (_length) {
			_s = new char[_length+1];
			strcpy(_s, other.c_str());
		} else {
			_s = NULL;
		}
		return *this;
	}
	String operator+(const String& o) const {
		if (!_length) return o;
		if (!o._length) return *this;
		String ret;
		ret._length = _length + o._length;
		ret._s = new char [ ret._length + 1];
		strcpy(ret._s, _s);
		strcpy(ret._s + _length, o._s);
		return ret;
	}
	String operator+(char c) const {
		String ret;
		ret._length = _length + 1;
		ret._s = new char[_length + 2];
		if (_length) strcpy(ret._s, _s);
		ret._s[_length] = c;
		ret._s[_length+1] = 0;
		return ret;
	}
	const String& operator+=(const String& o) {
		if (!_length) {
			*this = o;
			return *this;
		}
		if (!o._length) return *this;
		const char* old = _s;
		_s = new char [ _length + o._length + 1 ];
		strcpy(_s, old);
		strcpy(_s + _length, o._s);
		delete [] old;
		_length += o._length;
		return *this;
	}
	const String& operator+=(char c) {
		const char* old = _s;
		_s = new char[_length + 2];
		if (_length) {
			strcpy(_s, old);
			delete [] old;
		}
		_s[_length++] = c;
		_s[_length] = 0;
		return *this;
	}
	const char* c_str() const { return _s ? _s : ""; }
	String substr(size_t pos , size_t n = string::npos) const {
		if (pos > _length) {
			cerr << "out of bounds (length=" << _length << ", pos=" << pos << ")" << endl;
			abort();
		}
		if (n == string::npos or pos + n > _length)
			n = _length - pos;

		String ret;
		if (n) {
			ret._s = new char[n + 1];
			ret._length = n;
			strncpy(ret._s, _s + pos, n);
			ret._s[n] = 0;
		}
		return ret;
	}
	char operator[](size_t n) const {
		if (n >= _length) {
			cerr << "out of bounds (length=" << _length << ", pos=" << n << ")" << endl;
			abort();
		}
		return _s[n];
	}
	char& operator[](size_t n) {
		if (n >= _length) {
			cerr << "out of bounds (length=" << _length << ", pos=" << n << ")" << endl;
			abort();
		}
		return _s[n];
	}
	const String& assign(const char* s, size_t n) {
		if (_s) delete [] _s;
		if (!s or n == 0 or *s == '\0') {
			_s = NULL;
			_length = 0;
			return *this;
		}
		_length = strlen(s);
		if (n < _length) _length = n;
		char* t;
		t = _s = new char[_length+1];
		uint32_t l;
		for (l = _length; l; t++, s++, l--)
			*t = *s;
		*t = '\0';
		return *this;
	}
	bool operator==(const String& o) const {
		if (!_s) return o._s == NULL;
		if (!o._s) return false;
		return strcmp(_s, o._s) == 0;
	}
	bool operator!=(const String& o) const {
		if (!_s) return o._s != NULL;
		if (!o._s) return true;
		return strcmp(_s, o._s) != 0;
	}
	bool operator<(const String& o) const {
		if (!_s) return o._s != NULL;
		if (!o._s) return false;
		return strcmp(_s, o._s) < 0;
	}
	bool operator>(const String& o) const {
		if (!_s) return false;
		if (!o._s) return true;
		return strcmp(_s, o._s) > 0;
	}
	int compare(const String& o) const {
		if (!_s) return o._s ? -1 : 0;
		if (!o._s) return 1;
		return strcmp(_s, o._s);
	}
	bool empty() const { return _s == NULL; }
	void clear() { _length = 0; if (_s) { delete [] _s; _s = NULL; } }
	size_t find(char c, size_t pos = 0) const {
		if (pos >= _length) return string::npos;
		for (const char* s = _s + pos; *s; ++s)
			if (*s == c) return s - _s;
		return string::npos;
	}
	void resize(size_t n) {
		if (_length <= n) return;
		const char* old = _s;
		_length = n;
		_s = new char [ n + 1 ];
		strncpy(_s, old, n);
		_s[n] = 0;
		delete [] old;
	}
	void reserve(size_t) const {}
	size_t rfind(char c, size_t pos = string::npos) const {
		if (!_length) return string::npos;
		if (pos >= _length) pos = _length - 1;
		for (pos++; pos; pos--) {
			if (_s[pos - 1] == c) return pos - 1;
		}
		return string::npos;
	}

	String& erase(size_t pos) { resize(pos); return *this; }

private:
	char *_s;
	uint32_t _length;
};

inline ostream& operator<<(ostream& o, const String& s)
{
	return o << s.c_str();
}

inline istream& getline(istream& is, String& str)
{
	string s;
	istream& ret = getline(is, s);
	str = s;
	return ret;
}

#else

typedef string String;
#endif

// Usefull class when it's usefull for integers to be initialised to 0, such as when members of collections:
class Uint32_t
{
public:
	Uint32_t() : _i(0) {}
	Uint32_t(const uint32_t& i) : _i(i) {}
	operator const uint32_t& () const { return _i; }
	operator uint32_t& () { return _i; }
private:
	uint32_t _i;
};


class Int32_t
{
public:
	Int32_t() : _i(0) {}
	Int32_t(const int32_t& i) : _i(i) {}
	operator const int32_t& () const { return _i; }
	operator int32_t& () { return _i; }
private:
	int32_t _i;
};


class Double
{
public:
	Double() : _i(0.0) {}
	Double(const double& i) : _i(i) {}
	operator const double& () const { return _i; }
	operator double& () { return _i; }
private:
	double _i;
};


// Pointer defaulting to NULL
template <class T>
class Pointer
{
public:
	Pointer() : _p(NULL) {}
	Pointer(T* p) : _p(p) {}
	operator T*() const { return _p; }
	operator T* & () { return _p; }
private:
	T* _p;
};


namespace Util {

inline String
getLocalTime(const time_t& time)
{
	const char *c = ctime(&time);
	return String(c, strlen(c) - 1);
}


// like the perl version, if you know the number of fields you expect:
// fields is an array of pointers with NULL as the last element
// returns number of fields actually read, up to the size of fields[]
int split(const String& str, String *fields[], char separator = ' ');

// this is much slower:
inline vector<String>
split(const String& str, char separator = ' ') 
{
	vector<String> ret;
	uint64_t to;
	uint64_t from = 0;
	do {
		to = str.find(separator, from);
		if (to == string::npos) {
			ret.push_back(String(str, from));
			break;
		}
		ret.push_back(String(str, from, to-from));
		from = to + 1;
	} while (1);

	return ret;
}


void split(const String& str, const char separator, /* returned */ Vector<String>& fields);

int parseTabSeparatedFile(const char*& str, string *fields[]);

template <class T> bool
value(T &t, const String& s, std::ios_base& (*f)(std::ios_base&) = std::dec)
{
	std::istringstream iss(s.c_str());
	return !(iss >> f >> t).fail();
}
bool valueI(int32_t& i, const String& s);	// integer value
bool valueU(uint32_t& i, const String& s);	// unsigned value

template <class T> T max(const T &a, const T &b) { return a > b ? a : b; }
template <class T> T min(const T &a, const T &b) { return a < b ? a : b; }

vector<String> getCommandLineOptions(const map<String, bool*>& flags, const map<String, int*> intOptions, const map<String, String*>& stringOptions, int argc, char* argv[], void (*usage)());

String hms(time_t seconds);

String reverseComplement(const String&);
void reverseComplement(char* to, const char* from, uint32_t n);	// "to" must point to a valid address that has enough space
inline void reverseComplement(char* to, const char* from) { reverseComplement(to, from, strlen(from)); }

uint64_t getMemoryUsage();
inline double getMemoryUsageGB() { double a = (double) getMemoryUsage(); return a / 1073741824.0; }

void setThousandSeparator(ostream& o);

inline void printFasta(ostream& o, const String& seqName, const String& seq)
{
	o << ">" << seqName << endl;

	uint64_t len = seq.length();
	for (uint64_t i = 0; i < len; i += 80) {
		o << seq.substr(i, 80) << endl;
	}
}

void checkCreateDirectory(const char* name);
String getCurrentDirectory();


/* returns a random number from 0 to range-1

Random number generator should be initialised (seeded) with something like: srand(time(NULL));
If you run several times the same program in parallel for independent tests, make sure srand() is given a different seeds each time.
This means you should not run srand(time(NULL)) in parallel all within a second, since time(NULL) has seconds as units
*/
inline uint32_t
random(uint32_t range)
{
	if (range == 0) return 0;
	uint32_t max = uint32_t(RAND_MAX / range) * range;
	uint32_t random;
	do { random = rand(); } while (random >= max);
	return random % range;
}

inline bool
isNaN(double number) { return number != number; }

} // namespace Util


/* Simple class to allocate small pieces of memory tightly, but without any re-allocation
 * In DEBUG mode, it simply uses new and delete
 * Otherwise, it never returns any memory, so not suitable for active memory management
 * Also not suitable for multi-threading
 */
class SimpleAllocator
{
public:
	static const uint64_t _arraySize = 67108864; // this may be reduced to, say, 1048576 for valgrind
	static void initialise() { _numberBlocks = 1024; _blocks = new uint8_t* [_numberBlocks]; _currentBlock = 0; _next = _blocks[0] = new uint8_t[_arraySize]; }
	static void freeAll() { for (uint16_t i = 0; i <= _currentBlock; ++i) delete [] _blocks[i]; delete [] _blocks; }
	static uint8_t* allocate(uint32_t size);
	static void free(uint8_t* p) {
#ifdef DEBUG
		delete [] p;
#endif
	}

private:
	static uint8_t** _blocks;
	static uint32_t _numberBlocks;
	static uint32_t _currentBlock;
	static uint8_t* _next;
};


template <class T>
class Stack
{
public:
	Stack() : _arraySize(256), _number(0), _array(new T [_arraySize]) {}
	~Stack() { delete [] _array; }
	bool isEmpty() const { return _number == 0; }
	void push(T r) {
		if (_number == _arraySize) increaseSize();
		_array[_number] = r;
		_number++;
	}
	T top() {
		if (!_number) throw("Empty stack");
		return _array[_number-1];
	}
	T pop() {
		if (!_number) throw("Empty stack");
		_number--;
		return _array[_number];
	}
	uint64_t depth() const { return _number; }
	
private:
	void increaseSize();

	uint64_t _arraySize;
	uint64_t _number;
	T* _array;
};


template <class T>
void
Stack<T>::increaseSize()
{
	_arraySize *= 2;
	T* newArray = new T [_arraySize];
	for (uint64_t i = 0; i < _number; ++i) {
		newArray[i] = _array[i];
	}
	delete [] _array;
	_array = newArray;
}
#endif
