/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "Variant.h"
#include "Util.h"
#include "assert.h"
#include <fstream>
#include <map>
#include <limits>

static const GenomicPosition maxLengthForAlignment = 65535;

DnaSequence
Variant::applyVariants(const DnaSequence& original, const VariantSet& v, GenomicPosition posOffset, char direction)
{
	DnaSequence ret;
	list<Variant> variants(v);
	variants.sort();

	const GenomicPosition originalLength = original.length();
	GenomicPosition lastPos = 0;
	list<Variant>::iterator it = variants.begin();
	for (; it != variants.end(); ++it) {
		if (direction == 'u' and it->_direction == 'd') continue;
		if (direction == 'd' and it->_direction == 'u') continue;
		const GenomicPosition pos = it->position() - posOffset;
		if (pos < lastPos) {
			cerr << "Warning: variant at " << (pos+1) << " overlaps with previous one - ignored\n";
			continue;
		}

		if (lastPos < originalLength)
			ret += original.substr(lastPos, pos - lastPos);
		ret += it->newSeq();
		lastPos = pos + it->oldSeqLength();
	}

	if (lastPos < originalLength)
		ret += original.substr(lastPos);

	return ret;
}


String // = header
Variant::loadVariants(const Genome& genome, istream& in, /* output: */ VariantSet& variants)
{
	// fields:
	string seqName, positionS, oldSeqLenS, newSeq;
	string *fields[] = { &seqName, &positionS, &oldSeqLenS, &newSeq, NULL };
	String header;
	String* textP = &header;
	
	string line;
	uint32_t lineNo = 0;
	while (getline(in, line)) {
		lineNo++;
		if (line.length() == 0) continue;
		if (line[0] == '#') {
			if (textP) {
				*textP += line;
				*textP += "\n";
			}
			continue;
		}
		const char* l = line.c_str();
		int numFields  = Util::parseTabSeparatedFile(l, fields);
		if (numFields < 4) {
			cerr << " line " << lineNo << ": incorrect number of fields\n";
			throw(1);
		}
		GenomicPosition chromoPos;
		if (!Util::value<GenomicPosition>(chromoPos, positionS)) {
			cerr << " line " << lineNo << ": expected integer for position field\n";
			throw(2);
		}
		chromoPos--;	// convert 1-based chromoPos in file to zero-based
		GenomicPosition oldSeqLen;
		if (!Util::value<GenomicPosition>(oldSeqLen, oldSeqLenS)) {
			cerr << " line " << lineNo << ": expected integer for old sequence length field\n";
			throw(3);
		}

		// this is to make it backwards compatible:
		size_t p = seqName.rfind(':');
		char direction = '\0';
		if (p != string::npos) {
			if (p+1 < seqName.length())
				direction = seqName[p+1];
			seqName.erase(p);
		}

		const ChromosomeIndex chromoIndex = genome.chromoIndex(seqName);
		if (chromoIndex == Genome::SequenceNotFound) {
			cerr << "Sequence " << seqName << " not found in genome" << endl;
			abort();
		}
		const GenomicPosition pos = genome.position(chromoIndex, chromoPos);
		if (pos == invalidGenomicPosition) {
			textP = NULL;
		} else {
			variants.push_back(Variant(pos, oldSeqLen, DnaSequence(newSeq), direction));
			textP = & (variants.rbegin()->_text);
		}
	}

	return header;
}

typedef uint32_t matrix_element;

inline uint16_t max(matrix_element a, matrix_element b) { return a > b ? a : b; }
inline uint16_t flooredDiff(matrix_element a, matrix_element b) { return a > b ? a - b : 0; }

// The scores below are all unsigned:
static const matrix_element scoreMatch = 5;	// added
static const matrix_element penaltyMismatch = 4; // subtracted
static const matrix_element penaltyGapExtension = 4; // subtracted
static const matrix_element penaltyGapOpen = 16;	// heavily penalize new variants, to reduce their number


// Smith Waterman alignment
bool
Variant::findVariantsSW(const char* const aa, const uint16_t al,
		const char* const bb, const uint16_t bl,
		/* returned: */ list<Variant>& variants, const GenomicPosition posOffset)
{
	if (!al) {
		if (bl) {
			variants.push_back(Variant(posOffset, 0, DnaSequence(bb, bl)));
			return true; // - (int32_t) penaltyGapOpen - (int32_t) (bl - 1) * penaltyGapExtension;
		}
		return true; // 0;
	}
	if (!bl) {
		variants.push_back(Variant(posOffset, al, DnaSequence()));
		return true; //  - (int32_t) penaltyGapOpen - (int32_t) (al - 1) * penaltyGapExtension;
	}

	const uint32_t alp1 = al + 1;
	matrix_element *matrix;
	try {
		matrix = new matrix_element[ alp1 * (uint32_t) (bl+1)];
	} catch (bad_alloc) {
		cerr << "Memory allocation error trying to allocate: " << sizeof(matrix_element) * alp1 * (bl + 1) << " bytes" << endl;
		return false;
	}
#define MATRIX(i, j) matrix[(uint32_t) (i) + (uint32_t) (j) * alp1]
	uint16_t i, j, maxi = al, maxj = bl;
	matrix_element maxScore = 0;
//	int32_t totalScore = 0;

	// build the matrix:
	for (i = 0; i <= al; ++i) MATRIX(i, bl) = 0;
	for (j = 0; j < bl; ++j) MATRIX(al, j) = 0;
	for (i = al; i; --i) {
		for (j = bl; j; --j) {
			matrix_element score = 0;
			if (aa[i-1] == bb[j-1])
				score = MATRIX(i, j) + scoreMatch;
			else
				score = flooredDiff(MATRIX(i, j), penaltyMismatch);

			score = max(score, flooredDiff(MATRIX(i-1, j), penaltyGapExtension));
			score = max(score, flooredDiff(MATRIX(i, j-1), penaltyGapExtension));

			MATRIX(i-1, j-1) = score;
			if (score > maxScore) {
				maxScore = score;
				maxi = i-1;
				maxj = j-1;
			}
		}
	}

#if 0
	// print the matrix:
	for (j = 0; j < bl; j++) {
		cdbg << '\t' << bb[j];
	}
	cdbg << '\n';
	for (i = 0; i <= al; i++) {
		cdbg << (i == al ? ' ' : aa[i]) << '\t';
		for (j = 0; j < bl; j++) {
			cdbg << MATRIX(i, j) << '\t';
		}
		cdbg << MATRIX(i, bl) << '\n';
	}
	cdbg << "score = " << maxScore << " (" << maxi << ", " << maxj << ")\n";
#endif

	// read out the alignment:
	DnaSequence inserted;
	if (maxi > 0 or maxj > 0) {	// initial fragment not aligned
		uint16_t li = maxi, lj = maxj, i = 0;
		while (li and lj and aa[i] == bb[i]) { i++; li--; lj--; }
		while (li and lj and aa[i+li-1] == bb[i+lj-1]) { li--; lj--; }
		variants.push_back(Variant(posOffset + i, li, DnaSequence(bb + i, lj)));
/*		if (maxi == maxj) {
			totalScore -= (int32_t) maxi * penaltyMismatch;
		} else {
			totalScore -= (int32_t) penaltyGapOpen + (int32_t) (max(maxi, maxj) - 1) * penaltyGapExtension;
		}
*/
	}

	i = maxi;
	j = maxj;
//	if (i < al and j < bl) totalScore += scoreMatch;

	bool inVariant = false;
	uint16_t currentVariantI = 0, currentVariantJ = 0;
	// for current variant:
	uint16_t oldSeqLen = 0, newSeqLen = 0;

	while (i+1 < al and j+1 < bl) {
//		cdbg << i << ", " << j << ", v=" << variant;
		matrix_element penalty;
		matrix_element diag = MATRIX(i+1, j+1);
		if (inVariant) {
			penalty = 0;
		} else {
			penalty = penaltyGapOpen;
//			if (aa[i+1] != bb[j+1])
//				diag = flooredDiff(diag, penaltyNewMisMatch);
		}

		if (diag >= flooredDiff(MATRIX(i+1, j), penalty)) {
			if (diag >= flooredDiff(MATRIX(i, j+1), penalty)) {
				if (aa[i+1] == bb[j+1]) {
					if (inVariant) {
//						if (variant.oldSeqLength() != variant._newSeq.length())
//							totalScore -= (int32_t) (penaltyGapOpen - penaltyGapExtension);
						while (oldSeqLen and newSeqLen and aa[currentVariantI] == bb[currentVariantJ]) {
							currentVariantI++;
							currentVariantJ++;
							oldSeqLen--;
							newSeqLen--;
						}
						while (oldSeqLen and newSeqLen and aa[currentVariantI+oldSeqLen-1] == bb[currentVariantJ+newSeqLen-1]) {
							oldSeqLen--;
							newSeqLen--;
						}
//						cdbg << "NV=" << currentVariantI << '=' << string(aa + currentVariantI, oldSeqLen) << '=' << string(bb + currentVariantJ, newSeqLen) << endl;
						variants.push_back(Variant(posOffset + currentVariantI, oldSeqLen, DnaSequence(bb + currentVariantJ, newSeqLen)));
						oldSeqLen = newSeqLen = 0;
						inVariant = false;
					}
//					totalScore += scoreMatch;
				} else {
					if (!inVariant) {
//						cdbg << "variant starts here\n";
						inVariant = true;
						currentVariantI = i + 1;
						currentVariantJ = j + 1;
					}
//					totalScore -= (int32_t) penaltyMismatch;
					oldSeqLen++;
					newSeqLen++;
				}
				i++; j++;
			} else {
				if (!inVariant) {
//					cdbg << "variant starts\n";
					inVariant = true;
					currentVariantI = i + 1;
					currentVariantJ = j + 1;
				}
//				totalScore -= (int32_t) penaltyGapExtension;
				newSeqLen++;
				j++;
			}
		} else if (MATRIX(i+1, j) >= MATRIX(i, j+1)) {
			if (!inVariant) {
//				cdbg << "variant starts\n";
				inVariant = true;
				currentVariantI = i + 1;
				currentVariantJ = j + 1;
			}
//			totalScore -= (int32_t) penaltyGapExtension;
			oldSeqLen++;
			i++;
		} else {
			if (!inVariant) {
//				cdbg << "variant starts\n";
				inVariant = true;
				currentVariantI = i + 1;
				currentVariantJ = j + 1;
			}
//			totalScore -= (int32_t) penaltyGapExtension;
			newSeqLen++;
			j++;
		}
	}
	if (i+1 < al) {
		if (!inVariant) {
			inVariant = true;
			currentVariantI = i + 1;
			currentVariantJ = j + 1;
		}
		oldSeqLen += al - (i + 1);
//		totalScore -= (int32_t) penaltyGapExtension * (al - (i + 1));
	}
	if (j+1 < bl) {
		if (!inVariant) {
			inVariant = true;
			currentVariantI = i + 1;
			currentVariantJ = j + 1;
		}
		newSeqLen += bl - j - 1;
//		totalScore -= (int32_t) penaltyGapExtension * (bl - (j + 1));
	}
	if (inVariant) {
//		if (variant.oldSeqLength() != variant._newSeq.length())
//			totalScore -= (int32_t) (penaltyGapOpen - penaltyGapExtension);
		while (oldSeqLen and newSeqLen and aa[currentVariantI] == bb[currentVariantJ]) {
			currentVariantI++;
			currentVariantJ++;
			oldSeqLen--;
			newSeqLen--;
		}
		while (oldSeqLen and newSeqLen and aa[currentVariantI + oldSeqLen - 1] == bb[currentVariantJ + newSeqLen - 1]) {
			oldSeqLen--;
			newSeqLen--;
		}
//		cdbg << "NV=" << currentVariantI << '=' << string(aa + currentVariantI, oldSeqLen) << '=' << string(bb + currentVariantJ, newSeqLen) << endl;
		variants.push_back(Variant(posOffset + currentVariantI, oldSeqLen, DnaSequence(bb + currentVariantJ, newSeqLen)));
	}
	delete [] matrix;
	return true;
}


// input variants must be sorted
void
Variant::showAlignmentC(ostream& o, const char* aa, uint16_t al,
		const char* bb, uint16_t bl,
		const list<Variant>& variants, const GenomicPosition posOffset, const uint16_t lineLength)
{
	string a, x, b;
	GenomicPosition pos = posOffset;
	GenomicPosition l;
	list<Variant>::const_iterator it;
	for (it = variants.begin(); al and bl and it != variants.end(); ++it) {
		if (it->position() < pos) {
//			cdbg << "Variant at pos " << it->position() << " out of range (pos=" << pos << ")" << endl;
			continue;
		}

		for (l = it->position() - pos; l and al and bl; l--, aa++, bb++, al--, bl--, pos++) {
			a += *aa;
			x += '|';
			b += *bb;
		}
		GenomicPosition newSeqLen = it->newSeqLength();
		if (newSeqLen >= it->oldSeqLength()) {
			for (l = it->oldSeqLength(); l and al and bl; l--, aa++, bb++, al--, bl--, pos++) {
				a += *aa;
				x += ' ';
				b += *bb;
			}
			for (l = newSeqLen - it->oldSeqLength(); l and bl; l--, bb++, bl--) {
				a += '-';
				x += ' ';
				b += *bb;
			}
		} else {
			for (l = newSeqLen; l and al and bl; l--, aa++, bb++, al--, bl--, pos++) {
				a += *aa;
				x += ' ';
				b += *bb;
			}
			for (l = it->oldSeqLength() - newSeqLen; l and al; l--, aa++, al--, pos++) {
				a += *aa;
				x += ' ';
				b += '-';
			}
		}
	}

	if (al and bl) {
		for (l = min(al, bl); l; l--, aa++, bb++, al--, bl--) {
			a += *aa;
			x += '|';
			b += *bb;
		}
	}
	if (al) { 
		a += aa;
		x += string(al, ' ');
		b += string(al, '-');
	} else if (bl) {
		a += string(bl, '-');
		x += string(bl, ' ');
		b += bb;
	}
	if (lineLength) {
		for (GenomicPosition i = 0; i < a.length(); i += lineLength) {
			o << a.substr(i, lineLength) << endl << x.substr(i, lineLength) << endl
				<< b.substr(i, lineLength) << endl << endl;
		}
	} else {
		o << a << endl << x << endl << b << endl;
	}
}



void
Variant::showAlignment(ostream& o, const DnaSequence& a, const DnaSequence& b, const list<Variant>& variants, GenomicPosition posOffset, uint16_t lineLength)
{
	if (a.length() > maxLengthForAlignment or b.length() > maxLengthForAlignment) {
		cerr << "Variant::showAlignment sequences too long (" << a.length() << ", " << b.length() << ")\n";
		throw("Variant::showAlignment sequences too long");
	}
	return showAlignmentC(o, a.c_str(), a.length(), b.c_str(), b.length(), variants, posOffset, lineLength);
}


bool
Variant::parseAlignment(const String& aa, const String& bb, /* returned */ list<Variant>& variants)
{
	const GenomicPosition len = aa.length();
	if (len != bb.length()) {
		cerr << "Aligned sequences have different lengths: " << len << " vs " << bb.length() << endl;
		return false;
	}

	const char* a = aa.c_str();
	const char* b = bb.c_str();
	bool inVariant = false;
	Variant thisVariant;
	GenomicPosition apos = 0;
	for (GenomicPosition pos = 0; pos < len; ++pos, ++apos) {
		if (a[pos] == '-') {
			if (!inVariant) {
				inVariant = true;
				thisVariant.position() = apos;
			}
			thisVariant.newSeq() += b[pos];
			--apos;

		} else if (b[pos] == '-') {
			if (!inVariant) {
				inVariant = true;
				thisVariant.position() = apos;
			}
			thisVariant.oldSeqLength()++;

		} else if (a[pos] == b[pos]) {
			if (inVariant) {
				variants.push_back(thisVariant);
				thisVariant.clear();
				inVariant = false;
			}

		} else {
			if (!inVariant) {
				inVariant = true;
				thisVariant.position() = apos;
			}
			thisVariant.oldSeqLength()++;
			thisVariant.newSeq() += b[pos];
		}
	}

	if (inVariant)
		variants.push_back(thisVariant);

	return true;
}


int32_t
Variant::perfectMatchAlignmentScore(uint16_t length)
{
	return length * scoreMatch;
}


int32_t
Variant::alignmentScore(uint16_t lengthA, uint16_t lengthB, const list<Variant>& variants)
{
	int32_t score = 0;
	list<Variant>::const_iterator it;
	for (it = variants.begin(); it != variants.end(); ++it) {
		const GenomicPosition nsl = it->newSeqLength();
		if (it->oldSeqLength() == nsl) {
//			score -= (int32_t) penaltyNewMisMatch + (int32_t) (nsl - 1) * penaltyMismatch;
			score -= (int32_t) nsl * penaltyMismatch;
		} else if (it->oldSeqLength() > nsl) {
			score -= (int32_t) penaltyGapOpen + (int32_t) nsl * penaltyMismatch + (int32_t) (it->oldSeqLength() - nsl - 1) * penaltyGapExtension;
		} else {
			score -= (int32_t) penaltyGapOpen + (int32_t) it->oldSeqLength() * penaltyMismatch + (int32_t) (nsl - it->oldSeqLength() - 1) * penaltyGapExtension;
		}

		lengthA -= it->oldSeqLength();
		lengthB -= nsl;
	}
	assert(lengthA == lengthB);
	score += lengthA * scoreMatch;
	return score;
}


bool
Variant::findVariants(const DnaSequence& a, const DnaSequence& b, /* returned: */ list<Variant>& variants, GenomicPosition posOffset)
{
	if (a.length() > maxLengthForAlignment or b.length() > maxLengthForAlignment) {
		cerr << "Variant::findVariants sequences too long (" << a.length() << ", " << b.length() << ")\n";
		throw("Variant::findVariants sequences too long");
	}
	return findVariantsSW(a.c_str(), a.length(), b.c_str(), b.length(), variants, posOffset);
}


uint32_t
Variant::numberOfDifferences(const DnaSequence &seq1, const DnaSequence &seq2)
{
	list<Variant> variants;
	Variant::findVariants(seq1, seq2, variants);
	list<Variant>::iterator i;
	uint32_t count = 0;
	for (i = variants.begin(); i != variants.end(); ++i) {
		count += max(i->oldSeqLength(), i->newSeqLength());
	}
	return count;
}


uint32_t
Variant::numberOfDifferences(const list<Variant>& variants)
{
	list<Variant>::const_iterator i;
	uint32_t count = 0;
	for (i = variants.begin(); i != variants.end(); ++i) {
		count += max(i->oldSeqLength(), i->newSeqLength());
	}
	return count;
}


#if 0
 THIS NEEDS REVISION


string // = header
PrinsesVariant::loadVariants(istream& in, /* output: */ map<string, list<PrinsesVariant> >& variants)
{
	// fields:
	string seqName, positionS, oldSeqLenS, newSeq, contigInfo, scoreS, contigStartS, contigLengthS;
	string *fields[] = { &seqName, &positionS, &oldSeqLenS, &newSeq, &contigInfo, &scoreS, &contigStartS, &contigLengthS, NULL };
	string directionS, contigS, versionS;
	string *contigFields[] = { &directionS, &contigS, &versionS, NULL };
	string header;
	string* textP = &header;
	
	string line;
	uint32_t lineNo = 0;
	while (getline(in, line)) {
		lineNo++;
		if (line.length() == 0) continue;
		if (line[0] == '#') {
			*textP += line;
			*textP += "\n";
			continue;
		}
		const char* l = line.c_str();
		int numFields  = Util::parseTabSeparatedFile(l, fields);
		if (numFields < 4) {
			cerr << " line " << lineNo << ": incorrect number of fields\n";
			throw(1);
		}
		uint32_t position;
		if (!Util::value<uint32_t>(position, positionS)) {
			cerr << " line " << lineNo << ": expected integer for position field\n";
			throw(2);
		}
		position--;	// convert 1-based position in file to zero-based
		uint32_t oldSeqLen;
		if (!Util::value<uint32_t>(oldSeqLen, oldSeqLenS)) {
			cerr << " line " << lineNo << ": expected integer for old sequence length field\n";
			throw(3);
		}

		numFields = Util::split(contigInfo, contigFields, '.');
		uint32_t contigNumber = 0, versionNumber = 0;
		char direction = 'b';
		if (numFields == 3) {
			if (!Util::value<uint32_t>(contigNumber, contigS)) {
				cerr << " line " << lineNo << ": expected integer for contig number\n";
				throw(4);
			}
			if (!Util::value<uint32_t>(versionNumber, versionS)) {
				cerr << " line " << lineNo << ": expected integer for contig number\n";
				throw(5);
			}
			if (!directionS.empty())
				direction = directionS[0];
		}
		uint32_t averageScore = 0;
		Util::value<uint32_t>(averageScore, scoreS);
		uint64_t contigStart = 0;
		Util::value<uint64_t>(contigStart, contigStartS);
		uint32_t contigLength = 0;
		Util::value<uint32_t>(contigLength, contigLengthS);

		variants[seqName].push_back(PrinsesVariant(position, oldSeqLen, DnaSequence(newSeq), direction, contigNumber, versionNumber, averageScore, contigStart, contigLength));
		textP = & (variants[seqName].rbegin()->_text);
	}

	return header;
}
#endif
